from dataclasses import dataclass
from typing import TypedDict
from athina.loaders.loader import DataPoint


class LlmEvalResult(TypedDict):
    """
    A class to represent the result of an LLM evaluation.
    """

    name: str
    data: DataPoint
    failure: int
    reason: str
    runtime: int
    model: str



import asyncio
import pkg_resources
import requests
from typing import List, Optional
from athina.interfaces.athina import AthinaFilters, AthinaInference
from athina.keys import AthinaApiKey

# BASE_API_URL = "https://log.athina.ai"
BASE_API_URL = "http://localhost:9000"
SDK_VERSION = pkg_resources.get_distribution("athina-evals").version


class AthinaApiService:
    @staticmethod
    def _headers():
        if not AthinaApiKey.is_set():
            raise Exception(
                """Please sign up at https://athina.ai and set an Athina API key. 
                See https://docs.athina.ai/evals/quick_start for more information."""
            )
        athina_api_key = AthinaApiKey.get_key()
        return {
            "athina-api-key": athina_api_key,
        }

    @staticmethod
    def fetch_inferences(
        filters: Optional[AthinaFilters], limit: int = 50
    ) -> List[AthinaInference]:
        """
        Load data from Athina API.
        """
        raise NotImplementedError("This method has not been implemented yet.")

    @staticmethod
    def log_eval_results() -> List[AthinaInference]:
        """
        Logs eval results to Athina
        """
        pass

    @staticmethod
    def log_usage(evalName: str) -> List[AthinaInference]:
        """
        Logs a usage event to Posthog via Athina.
        """
        try:
            endpoint = f"{BASE_API_URL}/api/v1/sdk/log-usage"
            requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json={
                    "sdkVersion": SDK_VERSION,
                    "evalName": evalName,
                },
            )
        except Exception as e:
            # Silent failure is ok here.
            pass



import json


class JsonHelper:
    @staticmethod
    def _extract_json(data_string: str) -> str:
        """
        Extracts a JSON string from a larger string.
        Assumes the JSON content starts with '{' and continues to the end of the input string.
        """
        try:
            start_index = data_string.index("{")
            end_index = data_string.rfind("}")
            json_string = data_string[start_index : end_index + 1]
        except Exception as e:
            print("Failed to extract json", e)
            json_string = data_string
        return json_string

    @staticmethod
    def _load_json_from_text(text):
        """
        Extracts and loads a JSON string from a given text.
        """
        try:
            data = json.loads(text)
        except json.decoder.JSONDecodeError:
            data = None
        return data

    @staticmethod
    def extract_json_from_text(text):
        # In case you cannot handle an error, return None
        if text is None:
            return None
        response_json_format = JsonHelper._extract_json(text)
        response_json = JsonHelper._load_json_from_text(response_json_format)
        return response_json



from typing import List
from ..llm_evaluator import LlmEvaluator
from .examples import DOES_RESPONSE_ANSWER_QUERY_EVAL_EXAMPLES


class DoesResponseAnswerQuery(LlmEvaluator):
    """
    This evaluator checks if the response answers specifically what the user is asking about, and covers all aspects of the user's query.
    """

    SYSTEM_MESSAGE_TEMPLATE = """
    You are an expert at evaluating whether the response answers specifically what the user is asking about, and covers all aspects of the user's query.
    You are not checking for correctness, or factual accuracy. You are only checking if the response answers the user's query.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step.
        1. Consider the following: 
        user's query: {query}.
        response: {response}.
        2. Make sure to also consider these instructions: {additional_instructions}
        3. Determine if the response answers specifically what the user is asking about, and covers all aspects of the user's query.
        4. Provide a brief explanation of why the response does or does not answer the user's query sufficiently, labeled as 'explanation', leading up to a verdict (Pass/Fail) labeled as 'result'.
        5. Return a JSON object in the following format: "result": 'result', "explanation": 'explanation'.

        ### EXAMPLES ###
        Here's are some examples: 
        {examples}

        Now consider the following:
        user's query: {query}.
        response: {response}.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def name(self):
        return "does_response_answer_query"

    def display_name(self):
        return "Does Response Answer Query"

    def default_model(self):
        return "gpt-4-1106-preview"

    def required_args(self):
        return ["query", "response"]

    def examples(self):
        return DOES_RESPONSE_ANSWER_QUERY_EVAL_EXAMPLES

    def _user_message(
        self,
        query: str,
        response: str,
        additional_instructions: str = "",
        **kwargs,
    ) -> str:
        return self.USER_MESSAGE_TEMPLATE.format(
            query=query,
            response=response,
            additional_instructions=additional_instructions,
            examples=self._examples_str(),
        )



from typing import List, TypedDict
from athina.evals.llm.llm_evaluator import LlmEvaluator, LlmEvalResult
from athina.loaders.loader import DataPoint


class DataPointWithEvalResults(TypedDict):
    """A data point with its evaluation results."""

    data_point: DataPoint
    eval_results: List[LlmEvalResult]


class LlmEvaluatorDescription(TypedDict):
    """A description of an LLM evaluator."""

    name: str
    display_name: str


class LlmBatchEvalResult(TypedDict):
    """Result of running a batch of LLM evaluations."""

    results: List[DataPointWithEvalResults]
    evals: List[LlmEvaluatorDescription]
    total_runtime: float
    passed_evals: int
    failed_evals: int
    total_evals: int
    total_datapoints: int


class EvalRunner:
    @staticmethod
    def batch_eval_result(
        datapoints_with_eval_results: List[DataPointWithEvalResults],
        eval_descriptions: List[LlmEvaluatorDescription],
    ) -> LlmBatchEvalResult:
        """
        Calculate metrics for a batch of LLM evaluations.

        Args:
            datapoints_with_eval_results: A list of DataPointWithEvalResults objects.

        Returns:
            A LlmBatchEvalResult object.
        """
        total_runtime = 0
        passed_evals = 0
        failed_evals = 0
        total_evals = 0

        # Iterate through each DataPointWithEvalResults
        for datapoint in datapoints_with_eval_results:
            for eval_result in datapoint.get("eval_results", []):
                total_evals += 1
                total_runtime += eval_result.get("runtime", 0)

                # Counting passed and failed evaluations
                if eval_result.get("failure"):
                    failed_evals += 1
                else:
                    passed_evals += 1

        total_datapoints = len(datapoints_with_eval_results)

        return LlmBatchEvalResult(
            evals=eval_descriptions,
            total_runtime=total_runtime,
            passed_evals=passed_evals,
            failed_evals=failed_evals,
            total_evals=total_evals,
            total_datapoints=total_datapoints,
            results=datapoints_with_eval_results,
        )

    @staticmethod
    def run_batch(
        evals: List[LlmEvaluator],
        dataset: List[DataPoint],
    ) -> List[LlmBatchEvalResult]:
        """
        Run a suite of LLM evaluations against a dataset.

        Args:
            evals: A list of LlmEvaluator objects.
            data: A list of data points.

        Returns:
            A list of LlmBatchEvalResult objects.
        """
        datapoints_with_eval_results = []
        for datapoint in dataset:
            eval_results = []
            for evaluator in evals:
                try:
                    eval_result = evaluator.run(**datapoint)
                    eval_result.pop("data", None)
                    eval_results.append(eval_result)
                except Exception as e:
                    print(f"Error evaluating entry {datapoint}: {e}")
            datapoints_with_eval_results.append(
                {
                    "data_point": datapoint,
                    "eval_results": eval_results,
                }
            )
        eval_descriptions = list(
            map(
                lambda x: {
                    "name": x.name(),
                    "display_name": x.display_name(),
                },
                evals,
            )
        )
        return EvalRunner.batch_eval_result(
            datapoints_with_eval_results=datapoints_with_eval_results,
            eval_descriptions=eval_descriptions,
        )



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.services.athina_api_service import AthinaApiService
from .loader import Loader, DataPoint


class ResponseLoader(Loader):
    """
    This class is a data loader for evals that only evaluate the response.

    Attributes:
        col_response (str): The column name corresponding to the response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with responses.
    """

    def __init__(
        self,
        col_response="response",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_response = col_response
        self._raw_dataset = {}
        self._processed_dataset: List[DataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            KeyError: If mandatory columns (response) are missing in the raw dataset.
        """
        for raw_instance in self._raw_dataset:
            # Check for mandatory columns in raw_instance
            if self.col_response not in raw_instance:
                raise KeyError(f"'{self.col_response}' not found in provided data.")
            # Create a processed instance with mandatory fields
            processed_instance = {
                "response": raw_instance[self.col_response],
            }

            # Store the results
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: Optional[int] = None,
    ) -> List[DataPoint]:
        """
        Load data from Athina API.
        By default, this will fetch the last 50 inferences from the API.
        """
        athina_inferences = AthinaApiService.load_inferences(filters, limit=limit)
        self._raw_dataset = list(
            map(
                lambda x: {
                    "response": x.prompt_response,
                },
                athina_inferences,
            )
        )
        self.process()
        return self._processed_dataset



from dataclasses import dataclass
from typing import Optional, Dict, List


@dataclass
class AthinaInference:
    """Athina PromptRun class"""

    prompt_response: Optional[str]
    prompt_slug: Optional[str]
    language_model_id: Optional[str]
    environment: Optional[str]
    topic: Optional[str]
    customer_id: Optional[str]
    context: Optional[dict]
    user_query: Optional[str]


@dataclass
class AthinaFilters:
    prompt_slug: Optional[str]
    language_model_id: Optional[str]
    environment: Optional[str]
    topic: Optional[str]
    customer_id: Optional[str]



from .rag_loader import RagLoader
from .response_loader import ResponseLoader
from .loader import DataPoint, Loader, LoadFormat

__all__ = [
    "RagLoader",
    "ResponseLoader",
    "DataPoint",
    "Loader",
    "LoadFormat",
]



from typing import List
from ..llm_evaluator import LlmEvaluator


class CustomGrader(LlmEvaluator):
    """
    This evaluator checks if the response is correct according to a provided `grading_criteria`.
    """

    _examples = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "grading_criteria" not in kwargs:
            raise Exception("grading_criteria is required for CustomLlmEvaluator")
        if "examples" in kwargs:
            self._examples = kwargs["examples"]

    def name(self):
        return "custom_grader"

    def display_name(self):
        return "Custom"

    def default_model(self):
        return "gpt-4-1106-preview"

    def required_args(self):
        return ["response"]

    def examples(self):
        return self._examples



from .config import ConfigHelper
from .json import JsonHelper
from .kwparser import KeyValueAction
from .loader_helper import LoaderHelper
from .logger import logger
from .run_helper import RunHelper

__all__ = [
    "ConfigHelper",
    "JsonHelper",
    "KeyValueAction",
    "LoaderHelper",
    "logger",
    "RunHelper",
]



from typing import List
from ..llm_evaluator import LlmEvaluator
from .examples import FAITHFULNESS_EVAL_EXAMPLES


class Faithfulness(LlmEvaluator):
    """
    This evaluator checks if the response can be inferred using the information provided as context.
    """

    SYSTEM_MESSAGE_TEMPLATE = """ 
    You are an expert at evaluating whether the response can be inferred using ONLY the information provided as context.
    You are not concerned with factual correctness or accuracy. You are only determining whether the response can be inferred directly from the information provided as context.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step.
        1. Consider the following: 
        context: {context}.
        response: {response}.
        2. Make sure to also consider these instructions: {additional_instructions}
        3. Determine if the response can be inferred using ONLY the information provided in the context.
        4. Provide a brief explanation of why the response can or cannot be inferred purely from the context, labeled as 'explanation', leading up to a verdict (Pass/Fail) labeled as 'result'.
        5. Return a JSON object in the following format: "result": 'result', "explanation": 'explanation'.

        ### EXAMPLES ###
        Here's are some examples: 
        {examples}

        Now consider the following:
        context: {context}.
        response: {response}.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def name(self):
        return "faithfulness"

    def display_name(self):
        return "Faithfulness"

    def default_model(self):
        return "gpt-4-1106-preview"

    def required_args(self):
        return ["context", "response"]

    def examples(self):
        return FAITHFULNESS_EVAL_EXAMPLES

    def _user_message(
        self,
        context: str,
        response: str,
        additional_instructions: str = "",
        **kwargs,
    ) -> str:
        return self.USER_MESSAGE_TEMPLATE.format(
            context=context,
            response=response,
            additional_instructions=additional_instructions,
            examples=self._examples_str(),
        )



# athina_evals/evals/__init__.py
from athina.evals.llm.does_response_answer_query.evaluator import (
    DoesResponseAnswerQuery,
)
from athina.evals.llm.context_contains_enough_information.evaluator import (
    ContextContainsEnoughInformation,
)
from athina.evals.llm.faithfulness.evaluator import Faithfulness
from athina.evals.llm.llm_evaluator import LlmEvaluator
from athina.evals.llm.custom_grader.evaluator import CustomGrader

__all__ = [
    "DoesResponseAnswerQuery",
    "ContextContainsEnoughInformation",
    "Faithfulness",
    "CustomGrader",
]



from typing import List
from ..llm_evaluator import LlmEvaluator
from .examples import CONTEXT_CONTAINS_ENOUGH_INFORMATION_EXAMPLES


class ContextContainsEnoughInformation(LlmEvaluator):
    """
    This evaluator checks if the user's query can be answered using only the information in the context.
    """

    SYSTEM_MESSAGE_TEMPLATE = """
    You are an expert at evaluating whether a chatbot can answer a user's query using ONLY the information provided to you as context.
    You are not concerned with factual correctness or accuracy. You only care whether the context contains enough information to answer the user's query.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step:

        1. Consider the following: 
        user's query: {query}.
        context: {context}.
        2. Make sure to also consider these instructions: {additional_instructions}
        3. Determine if the chatbot can answer the user's query with nothing but the "context" information provided to you.
        4. Provide a brief explanation of why the context does or does not contain sufficient information, labeled as 'explanation', leading up to a verdict (Pass/Fail) labeled as 'result'.
        5. Return a JSON object in the following format: "verdict": 'result', "explanation": 'explanation'.

        Here's are some examples: 
        {examples}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def name(self):
        return "context_contains_enough_information"

    def display_name(self):
        return "Context Contains Enough Information"

    def default_model(self):
        return "gpt-4-1106-preview"

    def required_args(self):
        return ["query", "context"]

    def examples(self):
        return CONTEXT_CONTAINS_ENOUGH_INFORMATION_EXAMPLES

    def _user_message(
        self,
        query: str,
        context: str,
        additional_instructions: str = "",
        **kwargs,
    ) -> str:
        return self.USER_MESSAGE_TEMPLATE.format(
            query=query,
            context=context,
            additional_instructions=additional_instructions,
            examples=self.EXAMPLES,
        )



from enum import Enum


class Model(Enum):
    """
    Supported models for evaluations.
    """

    GPT35_TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4"
    GPT4_1106_PREVIEW = "gpt-4-1106-preview"

    @staticmethod
    def is_supported(model_name: str) -> bool:
        """
        Checks if the model is supported.
        """
        return model_name in [model.value for model in Model]

    @staticmethod
    def supports_json_mode(model_name: str) -> bool:
        """
        Checks if the model supports json mode.
        """
        JSON_MODE_SUPPORTED_MODELS = [Model.GPT4_1106_PREVIEW]
        return model_name in [model.value for model in JSON_MODE_SUPPORTED_MODELS]



from abc import ABC, abstractmethod
import time
from typing import List, Optional
from athina.interfaces.result import LlmEvalResult
from athina.interfaces.model import Model
from athina.llms.openai_service import OpenAiService
from athina.helpers.logger import logger
from athina.helpers.json import JsonHelper
from athina.loaders import DataPoint
from athina.keys import AthinaApiKey
from athina.errors.exceptions import NoAthinaApiKeyException
from athina.services.athina_api_service import AthinaApiService
from .example import FewShotExample


class LlmEvaluator(ABC):
    llm_service: OpenAiService
    grading_criteria: str

    TEMPERATURE = 0.0

    RETURN_FORMAT_INSTRUCTIONS = """
    Provide a brief explanation of why the response does or does not pass the grading criteria, labeled as 'explanation', leading up to a verdict (Pass/Fail) labeled as 'result'.

    You MUST return a JSON object in the following format: { "result": 'result', "explanation": 'explanation' }.
    """

    SYSTEM_MESSAGE_TEMPLATE = f""" 
    ### INSTRUCTIONS ###
    You are an expert at evaluating chatbot responses, according to some grading criteria.

    If it passes the grading criteria, then your result is Pass, otherwise it is Fail.
    """

    USER_MESSAGE_TEMPLATE = """
    ### GRADING CRITERIA ###
    {grading_criteria}

    ### EXAMPLES ###
    {examples}

    ### RESPONSE TO EVALUATE ###
    {response}
    """

    EXAMPLES: FewShotExample = []

    def __init__(
        self,
        model: Optional[str] = None,
        grading_criteria: Optional[str] = None,
    ):
        if not AthinaApiKey.is_set():
            raise NoAthinaApiKeyException()

        self.llm_service = OpenAiService()
        self.grading_criteria = grading_criteria if grading_criteria else ""
        if model is None:
            self.model = self.default_model()
        elif not Model.is_supported(model):
            raise ValueError(f"Unsupported model: {model}")
        else:
            self.model = model

    # Abstract properties
    @property
    @abstractmethod
    def name(self):
        """A unique name identifier for the evaluator."""
        pass

    @property
    @abstractmethod
    def display_name(self):
        """A display name for the evaluator."""
        pass

    @property
    @abstractmethod
    def required_args(self):
        """A list of required arguments for the evaluator."""
        pass

    @property
    @abstractmethod
    def default_model(self):
        """The default model for the evaluator."""
        pass

    @property
    @abstractmethod
    def examples(self):
        """A list of examples for the evaluator."""
        pass

    # Common methods
    def _examples_str(self) -> str:
        return "\n".join([str(example) for example in self.examples()])

    def _system_message(self, **kwargs) -> str:
        return self.SYSTEM_MESSAGE_TEMPLATE + self.RETURN_FORMAT_INSTRUCTIONS

    def _user_message(self, **kwargs) -> str:
        return self.USER_MESSAGE_TEMPLATE.format(
            examples=self._examples_str(),
            grading_criteria=self.grading_criteria,
            **kwargs,
        )

    def _prompt_messages(self, **kwargs) -> List[dict]:
        return [
            {
                "role": "system",
                "content": self._system_message(**kwargs),
            },
            {
                "role": "user",
                "content": self._user_message(**kwargs),
            },
        ]

    def _validate_args(self, **kwargs) -> None:
        for arg in self.required_args():
            if arg not in kwargs:
                raise ValueError(f"Missing required argument: {arg}")

    def run(self, **kwargs) -> LlmEvalResult:
        start_time = time.time()

        # Log usage to Athina for analytics
        AthinaApiService.log_usage(evalName=self.name())

        # Validate that correct args were passed
        self._validate_args(**kwargs)

        # Construct Prompt
        messages = self._prompt_messages(**kwargs)

        # Run the LLM Completion
        if Model.supports_json_mode(self.model):
            chat_completion_response = self.llm_service.json_completion(
                model=self.model,
                messages=messages,
                temperature=self.TEMPERATURE,
            )
        else:
            chat_completion_response = self.llm_service.chat_completion(
                model=self.model,
                messages=messages,
                temperature=self.TEMPERATURE,
            )

        # Extract JSON object from LLM response
        chat_completion_response_json = JsonHelper.extract_json_from_text(
            chat_completion_response
        )

        try:
            result = chat_completion_response_json["result"]
            explanation = chat_completion_response_json["explanation"]
            failure = result == "Fail"
        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)

        return LlmEvalResult(
            name=self.name(),
            data=kwargs,
            failure=failure,
            reason=explanation,
            runtime=eval_runtime_ms,
            model=self.model,
        )

    def _validate_batch_args(self, data: List[DataPoint]) -> bool:
        """
        Validates that each entry in the batch has all the required arguments.
        """
        for i, entry in enumerate(data):
            for arg in self.required_args:
                if arg not in entry:
                    raise ValueError(
                        f"Data at index {i} is missing required argument: {arg}"
                    )
        return True

    def _run_batch_generator(self, data: List[DataPoint]):
        """
        Generator function for running a batch of evaluations.
        Iterates over a dataset, and runs the evaluator on each entry.
        """
        for entry in data:
            try:
                yield self.run(**entry)
            except Exception as e:
                logger.error(f"Error evaluating entry {entry}: {e}")
                yield None

    def run_batch(self, data: List[DataPoint]) -> List[LlmEvalResult]:
        """
        Runs the evaluator on a batch of data.
        """
        self._validate_batch_args(data)
        eval_results = list(self._run_batch_generator(data))

        # Log evaluation results to Athina
        if AthinaApiKey.is_set():
            AthinaApiService.log_eval_results(eval_results)

        return eval_results



from typing import Optional
from .messages import NO_ATHINA_API_KEY, NO_OPENAI_API_KEY


class CustomException(Exception):
    def __init__(
        self, message: Optional[str] = None, extra_info: Optional[dict] = None
    ):
        self.message = message
        self.extra_info = extra_info
        super().__init__(self.message)

    def __str__(self):
        if self.extra_info:
            return f"{self.message} (Extra Info: {self.extra_info})"
        return self.message


class NoAthinaApiKeyException(CustomException):
    def __init__(self, message: str = NO_ATHINA_API_KEY):
        super().__init__(message)


class NoOpenAiApiKeyException(CustomException):
    def __init__(self, message: str = NO_OPENAI_API_KEY):
        super().__init__(message)



NO_ATHINA_API_KEY = """
Please sign up at https://athina.ai and set an Athina API key.

See https://docs.athina.ai/evals/quick_start for more info.
"""

NO_OPENAI_API_KEY = """
Please set an OpenAI API key.

See https://docs.athina.ai/evals/quick_start for more info.
"""
